/* user-action.c -- "run-as-user" interface for APACHE Action targets

	Copyright 1998,2005,2006,2007,2008
		by Mark E. Mallett, MV Communications, Inc.

	See the package's "LICENSE" file for terms.


Specify this program as the script to handle an "Action" for a
script type for an Apache web server.  It will run the referenced
object as the owner of the target file, subject to certain
security conditions.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <time.h>
#include <errno.h>

#include "mvform.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-cgi.h>

#include "user-action.h"


/* Local Definitions */

typedef					/* Holds a target definition */
  struct tgtdef {
    struct tgtdef	*tgt_nextP;	/* Ptr to next one */
    char		*tgt_extP;	/* Extension of target */
    char		*tgt_cmdP;	/* Command to run */
  }  TGTDEF;

typedef					/* Invoker ID item */
  struct iuid {
    struct iuid		*ii_nextP;	/* Next one */
    uid_t		ii_uid;		/* Acceptable invoker id */
  }  IUID;

/* External data referenced */

extern	BSTR	*GbsP;			/* General BSTR */
extern	BSTR	*GtkbsP;		/* BSTR for tokens */
extern	TKNS	*GtknsP;		/* Tokens pool */

/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	char	*DocumentRootP;		/* Path to document root */

    /* Needed to satisfy linker reference because we use some unrelated
       function in an external module.
    */
	char	Formdir[PATH_MAX +1];	/* Dir for this form */

/* Local routines and forward references */

TGTDEF *find_target PROTO( ( char *extP ) );
void	get_target_def PROTO( (TKNS *tknsP) );
void	get_valid_users PROTO( (TKNS *tknsP) );

/* Private data */

static	IUID	*IuidP;			/* Acceptable invoker ID list */
static	uid_t	Minuid = 100 ;		/* Reject UIDs lower than this */
static	uid_t	Maxuid = 30000 ;	/* Reject UIDs higher than this */
static	TGTDEF	*TdefP;			/* Head of target info list */

main( argc, argv )
	int		argc;
	char		**argv;
{
	int		i;
	char		*targetP;	/* Name of action target */
	char		*extP;		/* Ptr to filename extension */
	uid_t		realuid;
	TGTDEF		*tdefP;		/* Ptr to target definition block */
	IUID		*iuidP;		/* Invoker UID thing */
	struct stat	statblk;
	char		buf[BIGBUFSIZE +1];
	char		cmd[BIGBUFSIZE +1];
	gid_t		groups[2];

    if ( !formmisc_init() )
	exit( 0 );

    /* Read our control file */
    if ( !read_control_file() )
	exit( 1 );

    /* Check out invoker ID */
    if ( IuidP != NULL ) {
	realuid = getuid();
	for ( iuidP = IuidP ; iuidP != NULL; iuidP = iuidP->ii_nextP )
	    if ( realuid == iuidP->ii_uid )
		break;
	if ( iuidP == NULL ) {
	    /* Ran off the list without a match */
	    fprintf( stderr, "Invalid invoking UID (%d)\n", realuid );
	    exit( 1 );
	}
    }

    /* Get top of document tree */
    DocumentRootP = getenv( "DOCUMENT_ROOT" );
    if ( ( DocumentRootP == NULL ) || ( *DocumentRootP != '/' ) ) {
	fprintf( stderr, "Can't determine DOCUMENT_ROOT\n" );
	exit( 1 );
    }

    /* The path to the actual target script is in PATH_TRANSLATED */
    targetP = getenv( "PATH_TRANSLATED" );
    if ( ( targetP == NULL ) || ( *targetP != '/' ) ) {
	fprintf( stderr, "Can't find script name from PATH_TRANSLATED.\n" );
	exit( 1 );
    }

    /* Check out the file */
    if ( stat( targetP, &statblk ) < 0 ) {
	fprintf( stderr, "Can't stat target file \"%s\" -- errno %d\n",
		targetP, errno );
	exit( 1 );
    }

    /* UID must be reasonable */
    if ( ( statblk.st_uid < Minuid ) || ( statblk.st_uid > Maxuid ) ) {
	fprintf( stderr, "Bad UID (%d) in target file \"%s\"\n", 
		statblk.st_uid, targetP );
	exit( 1 );
    }

    /* Find the filename extension */
    extP = targetP + strlen( targetP ) - 1;
    for ( ; ; --extP ) {
	if ( ( *extP == '/' ) || ( extP == targetP ) ) {
	    fprintf( stderr, "Can't find extension in filename \"%s\"\n",
		    targetP );
	    exit( 1 );
	}
	if ( *extP == '.' )
	    break;
    }
    ++extP;				/* Skip the '.' */
    tdefP = find_target( extP );
    if ( tdefP == NULL ) {
	fprintf( stderr, "Can't find definition to handle file \"%s\"\n",
		targetP );
	exit( 1 );
    }

    /* Change into file owner before running the handler script */
    if ( setgid( statblk.st_gid ) < 0 ) {
	fprintf( stderr, "Can't change to group owner of \"%s\"\n", targetP  );
	exit( 1 );
    }

    /* Wipe out the groups list */
    groups[0] = statblk.st_gid ;
    groups[1] = 0;
    if ( setgroups( 1, &groups[0] ) < 0 ) {
	fprintf( stderr, "setgroups failed (%d)\n", errno );
	exit( 1 );
    }

    if ( setuid( statblk.st_uid ) < 0 ) {
	fprintf( stderr, "Can't change to owner of \"%s\"\n", targetP  );
	exit( 1 );
    }

    /* Set up the command name */
    if ( *(tdefP->tgt_cmdP) == '/' ) {
	/* Absolute path, use it */
	i = snprintf( "%s", sizeof(cmd), tdefP->tgt_cmdP );
    }
    else {
	i = snprintf( &cmd[0], sizeof(cmd),
		       "%s/%s",
		       DocumentRootP, tdefP->tgt_cmdP );
    }
    if ( i >= sizeof(cmd) ) {
	fprintf( stderr, "Buffer overflow\n" );
	exit( 1 );
    }

    execl( &cmd[0], &cmd[0], NULL );

    fprintf( stderr, "Can't exec \"%s\"\n", &cmd[0] );
    exit( 1 );
}
/*

*//* read_control_file()

	Scan the control file for this program

Accepts :


Returns :

	< value >	TRUE if OK
			FALSE if not (with error already generated)

*/

int
read_control_file NOARGLIST
{
	int		status;
	int		cC;		/* Character count */
	char		*strP;
	FILE		*fP;

    /* Try to open the definition file */
    fP = fopen( CTLFILE, "r" );
    if ( fP == NULL ) {
	fprintf( stderr, "Can't open control file \"%s\"\n", CTLFILE );
	return ( FALSE );
    }

    /* Read and parse the defs */
    for ( status = TRUE; status; ) {
	cC = fgetbtrec( fP, GbsP,
		 ((TXR_END_CRLF)|(TXR_END_CR)|(TXR_END_LF)),
		 NULL );
	if ( cC < 0 )			/* EOF */
	    break;

	/* Tokenize the line */
	tkline( GbsP, GtknsP );
	if ( GtknsP->tp_tC == 0 )	/* Blank line */
	    continue;

	/* Switch on first token. */
	strP = (char *)TKN_P( GtknsP, 0 );
	// if ( strcmp( strP, "#" ) == 0 )
	if ( *strP == '#' )
	    continue;			/* Comment */

	else if ( stricmp( strP, "Minimum-UID" ) == 0 ) {
	    /* Minimum allowable UID of target file */
	    if ( GtknsP->tp_tC > 1 )
		Minuid = atoi( (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Maximum-UID" ) == 0 ) {
	    /* Maximum allowable UID of target file */
	    if ( GtknsP->tp_tC > 1 )
		Maxuid = atoi( (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Target" ) == 0 ) {
	    get_target_def( GtknsP );
	}
	else if ( stricmp( strP, "Valid-User" ) == 0 ) {
	    get_valid_users( GtknsP );
	}
	else {
	    fprintf( stderr, "Bad token \"%s\" in control file.",
				strP );
	    status = FALSE;
	}
    }

    fclose( fP );

    return ( status );
}
/*

*//* get_target_def( tknsP )

	Scan a target definition 

Accepts :

	tknsP		Ptr to tokens pool

Returns :

	< nothing >

*/

void
get_target_def ARGLIST( (tknsP) )
    FARG( TKNS		*tknsP )	/* Tokens pool */
{
	int		tokX;
	char		*strP;
	char		*extP;		/* Extension */
	char		*commandP;	/* Commandfile */
	TGTDEF		*tdefP;		/* Target definition block */
	TGTDEF		**tdefPP;	/* Ptr ptr */

    extP = NULL;
    commandP = NULL;

    /* Scan the tokens */
    for ( tokX = 1; tokX < tknsP->tp_tC; ++tokX ) {
	strP = (char *)TKN_P( tknsP, tokX );
	if ( stricmp( strP, "command" ) == 0 ) {
	    if ( ( tokX +1 ) < tknsP->tp_tC )
		commandP = (char *)TKN_P( tknsP, tokX +1 );
	}
	else if ( stricmp( strP, "extension" ) == 0 ) {
	    if ( ( tokX +1 ) < tknsP->tp_tC )
		extP = (char *)TKN_P( tknsP, tokX +1 );
	}
	else
	    fprintf( stderr, "Unrecognized target keyword \"%s\"\n", 
			strP );
    }

    /* Make sure we have a complete definition */
    if ( ( extP == NULL ) ||
	 ( commandP == NULL ) ) {
	fprintf( stderr, "Incomplete target definition.\n" );
	return;
    }

    /* Make sure we don't duplicate another one */
    if ( find_target( extP ) != NULL ) {
	fprintf( stderr, "Duplicate target definition for extension \"%s\"\n",
		extP );
	return;
    }

    /* Make new block */
    tdefP = (TGTDEF *)emalloc( "get_target_def", "target block",
				sizeof(TGTDEF) );
    tdefP->tgt_extP = newstr( "get_target_def", "extension", extP );
    tdefP->tgt_cmdP = newstr( "get_target_def", "command", commandP );

    /* Link this in */
    for ( tdefPP = &TdefP; *tdefPP != NULL;
		tdefPP = &((*tdefPP)->tgt_nextP ) )
	if ( strcmp( tdefP->tgt_extP, (*tdefPP)->tgt_extP) < 0 )
	    break;

    tdefP->tgt_nextP = *tdefPP;
    *tdefPP = tdefP;
}
/*

*//* get_valid_users( tknsP )

	Get a list of valid invoker UIDs

Accepts :

	tknsP		Ptr to tokens pool

Returns :

	< nothing >

*/

void
get_valid_users ARGLIST( (tknsP) )
    FARG( TKNS		*tknsP )	/* Tokens pool */
{
	int		tokX;
	uid_t		uid;
	char		*strP;
	struct passwd	*pwdP;
	IUID		*iuidP;

    for ( tokX = 1; tokX < tknsP->tp_tC; ++tokX ) {
	strP = (char *)TKN_P( tknsP, 0 );
	if ( isdigit( strP[0] ) )
	    uid = atoi( strP );
	else {
	    pwdP = getpwnam( strP );
	    if ( pwdP == NULL ) {
		fprintf( stderr, "Unknown user \"%s\" in control file\n",
			strP );
		continue;
	    }
	    else
		uid = pwdP->pw_uid;
	}
	iuidP = (IUID *)emalloc( "read_control_file", "iuid block",
			   sizeof( IUID) );
	iuidP->ii_nextP = IuidP;
	IuidP = iuidP;
	iuidP->ii_uid = uid;
    }
}
/*

*//* find_target( extP )

	Find a target definition

Accepts :

	extP		Extension to look up

Returns :

	< value >	Ptr to TGTDEF block if found
			NULL if not found

*/

TGTDEF *
find_target ARGLIST( (extP) )
    FARG( char		*extP )		/* Extension to look up */
{
	TGTDEF		*tdefP;

    /* Scan the list */
    for ( tdefP = TdefP ; tdefP != NULL; tdefP = tdefP->tgt_nextP )
	if ( strcmp( extP, tdefP->tgt_extP ) == 0 )
	    break;

    return ( tdefP );
}
